#include "hist.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "symlist.h"

/* ultra_time: count the total number of hits in the histogram data.
 * Some hits may not be attributed to any symbol (in which
 * case total_time != ultra_time in the function self_time), it is
 * then probably a bug of xgprof. A warning is emitted
 * if this is the case.
 */
static unsigned long ultra_time;

int init_hist(hist *h,
    unsigned long lowpc, unsigned long highpc, unsigned long n)
{
  if (h->size) {
    fprintf(stderr, "only one hist handled for the moment.\n");
    exit(1);
  }
  h->data = malloc(n * sizeof(int));
  if (!h->data) return -1;

  h->lowpc = lowpc;
  h->highpc = highpc;
  h->size = n;

  return 0;
}

int add_hist(hist *h, int i, int v)
{
  h->data[i] = v;
  ultra_time+=v;
  return 0;
}

#if 0
static int count_sort_by_address(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  sym_info *sa = symbols + aa;
  sym_info *sb = symbols + bb;
  if (sa->lowpc < sb->lowpc) return -1;
  if (sa->lowpc > sb->lowpc) return 1;
  return 0;
}
#endif

int overlap(unsigned long min1, unsigned long max1,
            unsigned long min2, unsigned long max2)
{
  unsigned long min, max;
  if (max1 < min2) return -1;
  if (min1 > max2) return -2;
  if (min1 <= min2) min = min2; else min = min1;
  if (max1 >= max2) max = max2; else max = max1;
  return max-min+1;
}   

static int is_shared_bin(unsigned long min, unsigned long max, int sym,
                         int nb_syms)
{
  /* return 1 if previous or next symbol overlaps [min,max] */
  if (sym > 0) {
    /* not first symbol, check if overlaps with previous */
    if (overlap(min, max,
                symbols[sorted_symbols_by_address[sym-1]].lowpc,
                symbols[sorted_symbols_by_address[sym-1]].highpc) > 0)
      return 1;
  }

  if (sym < nb_syms - 1) {
    /* not last symbol, check if overlaps with next */
    if (overlap(min, max,
                symbols[sorted_symbols_by_address[sym+1]].lowpc,
                symbols[sorted_symbols_by_address[sym+1]].highpc) > 0)
      return 1;
  }

  return 0;
}

int self_time(hist *h)
{
  int cur_bin = 0;
  int cur_symbol = 0;
  int nb_syms = nb_unique_symbols();
  int cmp;
  unsigned long bin_low;
  unsigned long bin_high;
  unsigned long sym_low;
  unsigned long sym_high;
  unsigned long hist_bin_size = (h->highpc-h->lowpc)/h->size;
  double total_time = 0;

  fprintf(stderr, "histogram block size=%lu\n", hist_bin_size);

  while (cur_bin < h->size && cur_symbol < nb_syms) {
    bin_low = h->lowpc + hist_bin_size * cur_bin;
    bin_high = bin_low + hist_bin_size - 1;
    sym_low = symbols[sorted_symbols_by_address[cur_symbol]].lowpc;
    sym_high = symbols[sorted_symbols_by_address[cur_symbol]].highpc;
    cmp = overlap(bin_low, bin_high, sym_low, sym_high);
    if (cmp == -1) {
      /* bin < sym */
      cur_bin++;
    } else if (cmp == -2) {
      /* bin > sym */
      cur_symbol++;
    } else {
      /* overlap of cmp bytes, let's update internal time */
      double dtime;
      /* in case of partial overlap but symbols around don't take
       * any bit of the bin, give it all to the current symbol
       */
      if (cmp != hist_bin_size &&
          is_shared_bin(bin_low, bin_high, cur_symbol, nb_syms))
        dtime = (double)h->data[cur_bin] *
                ((double)cmp/(double)hist_bin_size);
      else
        dtime = h->data[cur_bin];
      total_time += dtime;
      symbols[sorted_symbols_by_address[cur_symbol]].internal_time += dtime;
      if (bin_high < sym_high)
        cur_bin++;
      else
        cur_symbol++;
    }
  }

  fprintf(stderr, "histogram total time %f/100 s\n", total_time);
  if (total_time != ultra_time) {
    fprintf(stderr, "histogram ultra time %lu/100 s\n", ultra_time);
    fprintf(stderr, "WARNING histogram ultra time != total time\n");
  }

  return 0;
}

typedef struct {
  list call;
  int nb_calls;
  list symbols;
  double internal_time;
  double cumuled_time_1;
  double cumuled_time_2;
  int visited:1;
} strong_component;

static strong_component *components;
static int components_size;

static int N;
static list L;

#if 0
static void visit(int node)
{
  sym_info *b = symbols + node;
  int i;
  int new;

  if (list_add(&L, node) == -1) {
    fprintf(stderr, "brutal OOM in tarjan/visit\n");
    exit(2);
  }

  b->dfsnum = N;
  b->low = N;
  N++;

  for (i = 0; i < list_size(&b->call); i+=2) {
    int b2 = list_get(&b->call, i);
    sym_info *q = symbols + b2;

    if (!q->inT) {
      q->inT = 1;
      //b->inT = 1;
      visit(b2);
      if (q->low < b->low) b->low = q->low;
    } else {
      if (q->dfsnum < b->low) b->low = q->dfsnum;
    }
  }

  if (b->low != b->dfsnum) return;

  do {
    new = list_get(&L, L.size-1);
    symbols[new].visited = 1;
    L.size--;
    if (list_add(&components[components_size].symbols, new) == -1) {
      fprintf(stderr, "ultra brutal OOM in tarjan/visit\n");
      exit(2);
    }
  } while (new != node);

  fprintf(stderr, "tarjan new component %d (%d elem)\n", components_size,
          list_size(&components[components_size].symbols));

  if (list_size(&components[components_size].symbols) != 1)
    for (i=0; i<list_size(&components[components_size].symbols); i++)
      fprintf(stderr, "tarjan   %s\n",
              symbol_name(list_get(&components[components_size].symbols, i)));

  components_size++;
}

/* algorithm from http://www.ics.uci.edu/~eppstein/161/960220.html */
/* does not work! (I guess I messed up somewhere or they did) */
static int tarjan(void)
{
  int i;
  int n = nb_symbols();
  sym_info *b;

  components = malloc(n * sizeof(strong_component));
  if (!components) return -1;
  memset(components, 0, n * sizeof(strong_component));

  /* do all symbols */
  for (i = 0; i < n; i++) {
    b = symbols + i;
    /* don't do unused symbols */
    if (!list_size(&b->call) && !list_size(&b->called_by) && !b->count)
      continue;
    if (b->visited) continue;
    visit(i);
  }

  return 0;
}
#endif /* #if 0 */

/* recursivity should be removed here (it is not very hard) */
static void visit2(int node)
{
  sym_info *b = symbols + node;
  int i;
  int new;

  b->low = N;
  b->dfsnum = N;
  N++;

  b->inT = 1;
  b->visited = 1;

  if (list_add(&L, node) == -1) {
    fprintf(stderr, "brutal OOM in tarjan/visit\n");
    exit(2);
  }

  for (i = 0; i < list_size(&b->call); i+=2) {
    int b2 = list_get(&b->call, i);
    sym_info *q = symbols + b2;

    if (!q->visited) {
      visit2(b2);
      if (q->low < b->low) b->low = q->low;
    } else if (q->inT) {
      if (q->dfsnum < b->low) b->low = q->dfsnum;
    }
  }

  if (b->low != b->dfsnum) return;

  do {
    new = list_get(&L, L.size-1);
    components[components_size].internal_time += symbols[new].internal_time;
    L.size--;
    symbols[new].inT = 0;
    symbols[new].scc = components_size;
    if (list_add(&components[components_size].symbols, new) == -1) {
      fprintf(stderr, "ultra brutal OOM in tarjan/visit\n");
      exit(2);
    }
  } while (new != node);

  fprintf(stderr, "tarjan new component %d (%d elem) (%f/100s)\n",
          components_size,
          list_size(&components[components_size].symbols),
          components[components_size].internal_time);

  if (list_size(&components[components_size].symbols) != 1)
    for (i=0; i<list_size(&components[components_size].symbols); i++)
      fprintf(stderr, "tarjan   %s\n",
              symbol_name(list_get(&components[components_size].symbols, i)));

  components_size++;
}

/* this one from
 http://en.wikipedia.org/wiki/Tarjan's_strongly_connected_components_algorithm
 */
static int tarjan2(void)
{
  int i;
  int n = nb_unique_symbols();
  sym_info *b;

  components = malloc(n * sizeof(strong_component));
  if (!components) return -1;
  memset(components, 0, n * sizeof(strong_component));

  /* do all symbols */
  for (i = 0; i < n; i++) {
    b = symbols + sorted_data[i];
    /* don't do unused symbols */
    if (!list_size(&b->call) && !list_size(&b->called_by) && !b->count)
      continue;
    if (b->visited) continue;
    visit2(sorted_data[i]);
  }

  return 0;
}

static void calc_scc_node(int node)
{
  int i;

  if (components[node].visited) return;

  components[node].visited = 1;

  components[node].cumuled_time_1 = components[node].internal_time;
  components[node].cumuled_time_2 = components[node].internal_time;

  for (i = 0; i < list_size(&components[node].call); i+=2) {
    calc_scc_node(list_get(&components[node].call, i));
#if 0
/* NOTE: this code is false, if this code is to be used again later in time */
    /* (too?) simple heuristic: a node gives 1/n of its time to each
     * of its n callers
     */
    components[node].cumuled_time +=
     components[list_get(&components[node].call, i)].cumuled_time /
     (double)list_size(&components[list_get(&components[node].call, i)].called);
#endif
    /* second try: be barbarian, give full time to all callers */
    components[node].cumuled_time_1 +=
      components[list_get(&components[node].call, i)].cumuled_time_1;

    /* let's be a bit more clever: # calls / # total calls of the scc */
    components[node].cumuled_time_2 +=
      components[list_get(&components[node].call, i)].cumuled_time_2 *
      (double)list_get(&components[node].call, i+1) /
      (double)components[list_get(&components[node].call, i)].nb_calls;
  }
}

static void scc_cumulate(void)
{
  int i;

  /* get the time of the children added up */
  for (i = 0; i < components_size; i++) {
    calc_scc_node(i);
  }
}

static int scc_list_add(list *l, int node, int nb_calls)
{
  int i;
  /* add the node if not already in the list
   * if node already there, simply sum up nb_calls
   */
  /* we could optimize to get O(log n) here instead of O(n)
   * but the lists are (normally) small
   */
  for (i = 0; i < list_size(l); i+=2)
    if (node == list_get(l, i)) {
      l->data[i+1] += nb_calls;
      return 0;
    }

  if (list_add(l, node) == -1) return -1;
  return list_add(l, nb_calls);
}

static int scc_call(int caller, int called, int nb_calls)
{
  if (scc_list_add(&components[caller].call, called, nb_calls) == -1)
    return -1;
  components[called].nb_calls += nb_calls;
//  if (scc_list_add(&components[called].called , caller, 0) == -1) return -1;
  return 0;
}

static int scc_build_call_graph(void)
{
  int i;
  int j;
  int n = nb_unique_symbols();
  int caller;
  int called;
  int nb_calls;

  for (i = 0; i < n; i++) {
    caller = symbols[sorted_data[i]].scc;
    for (j = 0; j < list_size(&symbols[sorted_data[i]].call); j+=2) {
      called = symbols[list_get(&symbols[sorted_data[i]].call, j)].scc;
      nb_calls = list_get(&symbols[sorted_data[i]].call, j+1);
      if (scc_call(caller, called, nb_calls) == -1) return -1;
    }
  }

  return 0;
}

static void scc_distribute_component(int cc)
{
  int i;
  strong_component *c = components + cc;

  for (i = 0; i < list_size(&c->symbols); i++) {
#if 0
    /* bad heuristic: distribute 1/n time to each member of scc */
    /* (it would be better to take into account # of calls of each member) */
    symbols[list_get(&c->symbols, i)].cumuled_time =
        c->cumuled_time / (double)list_size(&c->symbols);
#endif
    /* another try: give all time to each member */
    symbols[list_get(&c->symbols, i)].cumuled_time_1 = c->cumuled_time_1;

    /* distribute according to # calls of component / total # calls of scc */
    if ((double)symbols[list_get(&c->symbols, i)].count == (double)c->nb_calls)
      symbols[list_get(&c->symbols, i)].cumuled_time_2 = c->cumuled_time_2;
    else
      symbols[list_get(&c->symbols, i)].cumuled_time_2 = c->cumuled_time_2 *
        (double)symbols[list_get(&c->symbols, i)].count / (double)c->nb_calls;
  }
}

static void scc_distribute(void)
{
  int i;

  for (i = 0; i < components_size; i++) {
    scc_distribute_component(i);
  }
}

static int cumul_time(hist *h)
{
  /* calculate strongly connected components and getting their
   * self cumulated time
   */
  if (tarjan2() == -1) return -1;

  /* build call graph of components */
  if (scc_build_call_graph() == -1) return -1;

  /* cumulate times */
  scc_cumulate();

  /* distribute time to nodes */
  scc_distribute();

  return 0;
}

int calculate_hist(hist *h)
{
  int i;
  int n = nb_unique_symbols();

  if (h->size == 0) return 0;

  for (i=0; i<n; i++) {
    symbols[sorted_data[i]].lowpc = get_symbol_lowpc(sorted_data[i]);
    symbols[sorted_data[i]].highpc = get_symbol_highpc(sorted_data[i]);
  }

#if 0
  /* sort by address */
  /* it is useless but I leave the code anyway */
  sorted_symbols_by_address = malloc(n * sizeof(int));
  if (!sorted_symbols_by_address) return -1;
  for (i = 0; i < n; i++) sorted_symbols_by_address[i] = sorted_data[i];
  qsort(sorted_symbols_by_address, n, sizeof(int), count_sort_by_address);
#endif
  sorted_symbols_by_address = sorted_data;

  if (self_time(h) == -1) return -1;
  if (cumul_time(h) == -1) return -1;

  return 0;
}
